/*-
 * Copyright (c) 2016 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache.h"
#include "vrt.h"
#include "vcl.h"
#include "vas.h"
#include "vsb.h"
#include "miniobj.h"

#include "vcc_if.h"
#include "vre2/vre2.h"
#include "vre2/vre2set.h"

#define ERR(ctx, msg) \
	errmsg((ctx), "vmod re2 error: " msg)

#define VERR(ctx, fmt, ...) \
	errmsg((ctx), "vmod re2 error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
	VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define INIT(ctx) (((ctx)->method & VCL_MET_INIT) != 0)

struct vmod_re2_regex {
	unsigned	magic;
#define VMOD_RE2_REGEX_MAGIC 0x5c3f6f24
	vre2		*vre2;
	char		*vcl_name;
	int		ngroups;
	unsigned	never_capture;
};

struct vmod_re2_set {
	unsigned	magic;
#define VMOD_RE2_SET_MAGIC 0xf6d7b15a
	vre2set		*set;
	char		*vcl_name;
	char		**string;
	VCL_BACKEND	*backend;
	unsigned	compiled;
	int		npatterns;
};

typedef struct task_match_t {
	unsigned	magic;
#define TASK_MATCH_MAGIC 0xa4b93c57
	vre2		*vre2;
	void		*groups;
	int		ngroups;
	unsigned	never_capture;
} task_match_t;

struct task_set_match {
	unsigned	magic;
#define TASK_SET_MATCH_MAGIC 0x7a24a90b
	int		*matches;
	size_t		nmatches;
};

static char c;
static const void *match_failed = (void *) &c;

static size_t match_sz;

static const char * const rewrite_name[] = {
	[SUB]     = "sub",
	[SUBALL]  = "suball",
	[EXTRACT] = "extract",
};

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		VSLv(SLT_VCL_Error, 0, fmt, args);
	va_end(args);

	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		va_start(args, fmt);
		VSB_vprintf(ctx->msg, fmt, args);
		VSB_putc(ctx->msg, '\n');
		va_end(args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
}

static void
free_task_match(void *p)
{
	task_match_t *task_match;
	CAST_OBJ_NOTNULL(task_match, p, TASK_MATCH_MAGIC);
	if (task_match->vre2 != NULL)
		vre2_fini(&task_match->vre2);
}

#define ERR_PREFIX "match(\"%.40s\"): "

static VCL_BOOL
match(VRT_CTX, vre2 * restrict vre2, VCL_STRING restrict subject,
      void * restrict * const groups, const unsigned never_capture,
      int ngroups)
{
	int match = 0, len;
	const char *err;
	char *text = (void *) subject;
	uintptr_t snap;

	AN(groups);
	if (subject == NULL)
		subject = "";
	len = strlen(subject);
	*groups = NULL;

	snap = WS_Snapshot(ctx->ws);
	if (!never_capture) {
		ngroups++;
		if (!WS_Inside(ctx->ws, subject, subject + len + 1))
			if ((text = WS_Copy(ctx->ws, subject, len + 1))
			    == NULL) {
				VERR(ctx, ERR_PREFIX "insufficient workspace "
				     "to copy subject", subject);
				return 0;
			}
		if ((*groups = WS_Alloc(ctx->ws, ngroups * match_sz))
		    == NULL) {
			VERR(ctx, ERR_PREFIX "insufficient workspace to "
			     "allocate match data", subject);
			WS_Reset(ctx->ws, snap);
			return 0;
		}
	}

	if ((err = vre2_match(vre2, text, len, &match, ngroups, *groups))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "%s", subject, err);
		if (!never_capture)
			WS_Reset(ctx->ws, snap);
		return 0;
	}

	if (!match) {
		*groups = (void *)match_failed;
		if (!never_capture)
			WS_Reset(ctx->ws, snap);
	}
	return match;
}

#undef ERR_PREFIX

#define ERR_PREFIX "backref %ld, fallback \"%.40s\": "

static VCL_STRING
backref(VRT_CTX, VCL_INT refnum, VCL_STRING fallback, void * const groups,
	const int ngroups)
{
	const char *err, *capture;
	char *backref;
	int len;

	AN(groups);
	if (groups == match_failed)
		return fallback;

	WS_Assert_Allocated(ctx->ws, groups, (ngroups + 1) * match_sz);

	if ((err = vre2_capture(groups, (int) refnum, &capture, &len))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "error retrieving capture: %s", refnum,
		     fallback, err);
		return fallback;
	}

	assert(len >= 0);
	if (capture == NULL)
		return fallback;
	if (len == 0)
		return "";

	WS_Assert_Allocated(ctx->ws, capture, len);
	if ((backref = WS_Copy(ctx->ws, capture, len + 1)) == NULL) {
		VERR(ctx, ERR_PREFIX "insufficient workspace for backref",
		     refnum, fallback);
		return fallback;
	}
	backref[len] = '\0';
	return backref;
}

#undef ERR_PREFIX

#define ERR_PREFIX "namedref name=\"%.40s\", fallback=\"%.40s\": "

static VCL_STRING
namedref(VRT_CTX, vre2 * restrict vre2, VCL_STRING name, VCL_STRING fallback,
	 void * const restrict groups, const int ngroups)
{
	int refnum;
	const char *err;

	if ((err = vre2_get_group(vre2, name, &refnum)) != NULL) {
		VERR(ctx, ERR_PREFIX "%s", name, fallback, err);
		return fallback;
	}
	if (refnum == -1) {
		VERR(ctx, ERR_PREFIX "no such named group", name, fallback);
		return fallback;
	}
	assert(refnum > 0 && refnum <= ngroups);
	return backref(ctx, refnum, fallback, groups, ngroups);
}

#undef ERR_PREFIX

#define ERR_PREFIX \
	"%s(text=\"%.40s\", rewrite=\"%.40s\", fallback=\"%.40s\"): "

static VCL_STRING
rewritef(VRT_CTX, vre2 * restrict vre2, const rewrite_e mode, VCL_STRING text,
	 VCL_STRING rewrite, VCL_STRING fallback)
{
	int match = 0;
	size_t bytes, len;
	char *ret;
	const char *err;

	ret = WS_Front(ctx->ws);
	bytes = WS_Reserve(ctx->ws, 0);
	if (bytes == 0) {
		VERR(ctx, ERR_PREFIX "workspace overflow", rewrite_name[mode],
		     text, rewrite, fallback);
		return fallback;
	}
	if ((err = vre2_rewrite(vre2, mode, text, rewrite, ret, bytes, &match,
				&len))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "%s", rewrite_name[mode], text, rewrite,
		     fallback, err);
		WS_Release(ctx->ws, 0);
		return fallback;
	}
	if (!match) {
		WS_Release(ctx->ws, 0);
		return fallback;
	}
	WS_Release(ctx->ws, len);
	return ret;
}

#undef ERR_PREFIX

/* Event function */

int
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) ctx;
	(void) priv;

	if (e == VCL_EVENT_LOAD)
		match_sz = vre2_matchsz();
	return 0;
}

/* Object regex */

VCL_VOID
vmod_regex__init(const struct vrt_ctx *ctx, struct vmod_re2_regex **rep,
		 const char *vcl_name, VCL_STRING pattern, VCL_BOOL utf8,
		 VCL_BOOL posix_syntax, VCL_BOOL longest_match, VCL_INT max_mem,
		 VCL_BOOL literal, VCL_BOOL never_nl, VCL_BOOL dot_nl,
		 VCL_BOOL never_capture, VCL_BOOL case_sensitive,
		 VCL_BOOL perl_classes, VCL_BOOL word_boundary,
		 VCL_BOOL one_line)
{
	struct vmod_re2_regex *re;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(rep);
	AZ(*rep);
	AN(vcl_name);
	AN(pattern);
	ALLOC_OBJ(re, VMOD_RE2_REGEX_MAGIC);
	AN(re);
	*rep = re;

	if ((err = vre2_init(&re->vre2, pattern, utf8, posix_syntax,
			     longest_match, max_mem, literal, never_nl, dot_nl,
			     never_capture, case_sensitive, perl_classes,
			     word_boundary, one_line)) != NULL) {
		VERR(ctx, "Cannot compile '%.40s' in %s constructor: %s",
		     pattern, vcl_name, err);
		return;
	}
	if (! never_capture) {
		if ((err = vre2_ngroups(re->vre2, &re->ngroups)) != NULL) {
			VERR(ctx, "Cannot obtain number of capturing groups in "
			     "%s constructor: %s", vcl_name, err);
			return;
		}
		assert(re->ngroups >= 0);
	}
	re->never_capture = never_capture;
	re->vcl_name = strdup(vcl_name);
}

VCL_VOID
vmod_regex__fini(struct vmod_re2_regex **rep)
{
	struct vmod_re2_regex *re;

	if (rep == NULL || *rep == NULL)
		return;
	CHECK_OBJ(*rep, VMOD_RE2_REGEX_MAGIC);
	re = *rep;
	*rep = NULL;
	vre2_fini(&re->vre2);
	if (re->vcl_name != NULL)
		free(re->vcl_name);
	FREE_OBJ(re);
}

VCL_BOOL
vmod_regex_match(const struct vrt_ctx *ctx, struct vmod_re2_regex *re,
		 VCL_STRING subject)
{
	VCL_BOOL ret;
	struct vmod_priv *task;
	task_match_t *task_match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);

	task = VRT_priv_task(ctx, re);
	AN(task);
	if (task->priv == NULL) {
		if ((task->priv = WS_Alloc(ctx->ws, sizeof(*task_match)))
		    == NULL) {
			VERRNOMEM(ctx, "Allocating match data in "
				  "%s.match(subject=\"%.40s\")",
				  re->vcl_name, subject);
			return 0;
		}
		task->len = sizeof(*task_match);
		task->free = NULL;
		task_match = (task_match_t *)task->priv;
		task_match->magic = TASK_MATCH_MAGIC;
	}
	else {
		WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*task_match));
		CAST_OBJ(task_match, task->priv, TASK_MATCH_MAGIC);
	}

	ret = match(ctx, re->vre2, subject, &task_match->groups,
		    re->never_capture, re->ngroups);
	return ret;
}

#define ERR_PREFIX "%s.backref(ref=%ld, fallback=\"%.40s\"): "

VCL_STRING
vmod_regex_backref(VRT_CTX, struct vmod_re2_regex *re, VCL_INT refnum,
		   VCL_STRING fallback)
{
	struct vmod_priv *task;
	task_match_t *task_match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	assert(refnum >= 0);
	if (fallback == NULL) {
		VERR(ctx, ERR_PREFIX "fallback is undefined", re->vcl_name,
		     refnum, "<undefined>");
		return "**BACKREF METHOD FAILED**";
	}

	if (re->never_capture) {
		VERR(ctx, ERR_PREFIX "never_capture is true for object %s",
		     re->vcl_name, refnum, fallback, re->vcl_name);
		return fallback;
	}
	if (refnum > re->ngroups) {
		VERR(ctx, ERR_PREFIX "backref out of range (max %d)",
		     re->vcl_name, refnum, fallback, re->ngroups);
		return fallback;
	}

	task = VRT_priv_task(ctx, re);
	AN(task);
	if (task->priv == NULL) {
		VERR(ctx, ERR_PREFIX "backref called without prior match",
		     re->vcl_name, refnum, fallback);
		return fallback;
	}
	WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*task_match));
	CAST_OBJ(task_match, task->priv, TASK_MATCH_MAGIC);
	return backref(ctx, refnum, fallback, task_match->groups, re->ngroups);
}

#undef ERR_PREFIX

#define ERR_PREFIX "%s.namedref(name=\"%.40s\", fallback=\"%.40s\"): "

VCL_STRING
vmod_regex_namedref(VRT_CTX, struct vmod_re2_regex *re, VCL_STRING name,
		    VCL_STRING fallback)
{
	struct vmod_priv *task;
	task_match_t *task_match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	if (fallback == NULL) {
		VERR(ctx, "%s.namedref(): fallback is undefined", re->vcl_name);
		return "**NAMEDREF METHOD FAILED**";
	}
	if (name == NULL || name[0] == '\0') {
		VERR(ctx, ERR_PREFIX "name is empty", re->vcl_name, "",
		     fallback);
		return fallback;
	}

	if (re->never_capture) {
		VERR(ctx, ERR_PREFIX "never_capture is true for object %s",
		     re->vcl_name, name, fallback, re->vcl_name);
		return fallback;
	}

	task = VRT_priv_task(ctx, re);
	AN(task);
	if (task->priv == NULL) {
		VERR(ctx, ERR_PREFIX "namedref called without prior match",
		     re->vcl_name, name, fallback);
		return fallback;
	}
	WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*task_match));
	CAST_OBJ(task_match, task->priv, TASK_MATCH_MAGIC);
	return namedref(ctx, re->vre2, name, fallback, task_match->groups,
			re->ngroups);
}

#undef ERR_PREFIX

static VCL_STRING
rewrite_method(VRT_CTX, const rewrite_e mode,
	       struct vmod_re2_regex * restrict re, VCL_STRING restrict text,
	       VCL_STRING restrict rewrite, VCL_STRING restrict fallback)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE2_REGEX_MAGIC);
	if (fallback == NULL) {
		VERR(ctx, "%s.%s(): fallback is undefined", re->vcl_name,
		     rewrite_name[mode]);
		switch(mode) {
		case SUB:
			return "**SUB METHOD FAILED**";
		case SUBALL:
			return "**SUBALL METHOD FAILED**";
		case EXTRACT:
			return "**EXTRACT METHOD FAILED**";
		default:
			WRONG("illegal mode");
		}
	}
	if (text == NULL) {
		VERR(ctx, "%s.%s(text=<undefined>, fallback=\"%.40s\"): "
		     "text is undefined", re->vcl_name, rewrite_name[mode],
		     fallback);
		return fallback;
	}
	if (rewrite == NULL) {
		VERR(ctx, "%s.%s(text=\"%.40s\", rewrite=<undefined>, "
		     "fallback=\"%.40s\"): rewrite is undefined", re->vcl_name,
		     rewrite_name[mode], text, fallback);
		return fallback;
	}

	return rewritef(ctx, re->vre2, mode, text, rewrite, fallback);
}

VCL_STRING
vmod_regex_sub(VRT_CTX, struct vmod_re2_regex *re, VCL_STRING text,
	       VCL_STRING rewrite, VCL_STRING fallback)
{
	return rewrite_method(ctx, SUB, re, text, rewrite, fallback);
}

VCL_STRING
vmod_regex_suball(VRT_CTX, struct vmod_re2_regex *re, VCL_STRING text,
		  VCL_STRING rewrite, VCL_STRING fallback)
{
	return rewrite_method(ctx, SUBALL, re, text, rewrite, fallback);
}

VCL_STRING
vmod_regex_extract(VRT_CTX, struct vmod_re2_regex *re, VCL_STRING text,
		   VCL_STRING rewrite, VCL_STRING fallback)
{
	return rewrite_method(ctx, EXTRACT, re, text, rewrite, fallback);
}

/* Object set */

VCL_VOID
vmod_set__init(VRT_CTX, struct vmod_re2_set **setp, const char *vcl_name,
	       VCL_ENUM anchor, VCL_BOOL utf8, VCL_BOOL posix_syntax,
	       VCL_BOOL longest_match, VCL_INT max_mem, VCL_BOOL literal,
	       VCL_BOOL never_nl, VCL_BOOL dot_nl, VCL_BOOL case_sensitive,
	       VCL_BOOL perl_classes, VCL_BOOL word_boundary, VCL_BOOL one_line)
{
	struct vmod_re2_set *set;
	anchor_e anchor_e;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(setp);
	AZ(*setp);
	AN(vcl_name);
	AN(anchor);
	ALLOC_OBJ(set, VMOD_RE2_SET_MAGIC);
	AN(set);
	*setp = set;

	if (strcmp(anchor, "none") == 0)
		anchor_e = NONE;
	else if (strcmp(anchor, "start") == 0)
		anchor_e = START;
	else if (strcmp(anchor, "both") == 0)
		anchor_e = BOTH;
	else
		WRONG("illegal anchor");

	if ((err = vre2set_init(&set->set, anchor_e, utf8, posix_syntax,
				longest_match, max_mem, literal, never_nl,
				dot_nl, case_sensitive, perl_classes,
				word_boundary, one_line))
	    != NULL) {
		VERR(ctx, "new %s = re2.set(): %s", vcl_name, err);
		return;
	}
	set->vcl_name = strdup(vcl_name);
	AZ(set->string);
	AZ(set->backend);
	AZ(set->npatterns);
}

VCL_VOID
vmod_set__fini(struct vmod_re2_set **setp)
{
	struct vmod_re2_set *set;

	if (setp == NULL || *setp == NULL)
		return;
	CHECK_OBJ(*setp, VMOD_RE2_SET_MAGIC);
	set = *setp;
	*setp = NULL;
	vre2set_fini(&set->set);
	if (set->vcl_name != NULL)
		free(set->vcl_name);
	FREE_OBJ(set);
}

#define ERR_PREFIX "%s.add(\"%.40s\"): "

VCL_VOID
vmod_set_add(VRT_CTX, struct vmod_re2_set *set, VCL_STRING pattern,
	     VCL_STRING string, VCL_BACKEND backend)
{
	const char *err;
	int n;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (pattern == NULL)
		pattern = "";
	if (!INIT(ctx)) {
		VERR(ctx, ERR_PREFIX ".add() may only be called in vcl_init",
		     set->vcl_name, pattern);
		return;
	}

	if (set->compiled) {
		VERR(ctx, ERR_PREFIX "%s has already been compiled",
		     set->vcl_name, pattern, set->vcl_name);
		return;
	}
	if ((err = vre2set_add(set->set, pattern, &n)) != NULL) {
		VERR(ctx, ERR_PREFIX "Cannot compile '%.40s': %s",
		     set->vcl_name, pattern, pattern, err);
		return;
	}

	if (string != NULL) {
		if ((set->string = realloc(set->string,
					   (n + 1) * (sizeof(char *))))
		    == NULL) {
			VERRNOMEM(ctx, ERR_PREFIX "adding string %s",
				  set->vcl_name, pattern, string);
			return;
		}
		set->string[n] = strdup(string);
	}
	if (backend != NULL) {
		if ((set->backend = realloc(set->backend,
					    (n + 1) * (sizeof(VCL_BACKEND))))
		    == NULL) {
			VERRNOMEM(ctx, ERR_PREFIX "adding backend %s",
				  set->vcl_name, pattern,
				  VRT_BACKEND_string(backend));
			return;
		}
		set->backend[n] = backend;
	}
	set->npatterns++;
}

#undef ERR_PREFIX

#define ERR_PREFIX "%s.compile(): "

VCL_VOID
vmod_set_compile(VRT_CTX, struct vmod_re2_set *set)
{
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (!INIT(ctx)) {
		VERR(ctx, ERR_PREFIX ".compile() may only be called in "
		     "vcl_init", set->vcl_name);
		return;
	}
	if (set->npatterns == 0) {
		VERR(ctx, ERR_PREFIX "no patterns were added", set->vcl_name);
		return;
	}

	if (set->compiled) {
		VERR(ctx, ERR_PREFIX "%s has already been compiled",
		     set->vcl_name, set->vcl_name);
		return;
	}
	if ((err = vre2set_compile(set->set)) != NULL) {
		VERR(ctx, ERR_PREFIX "failed, possibly insufficient memory",
		     set->vcl_name);
		return;
	}
	set->compiled = 1;
}

#undef ERR_PREFIX

#define ERR_PREFIX "%s.match(\"%.40s\"): "

VCL_BOOL
vmod_set_match(VRT_CTX, struct vmod_re2_set *set, VCL_STRING subject)
{
	int match = 0;
	struct vmod_priv *priv;
	struct task_set_match *task;
	char *buf;
	size_t buflen;
	const char *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (subject == NULL)
		subject = "";

	if (!set->compiled) {
		VERR(ctx, ERR_PREFIX "%s was not compiled", set->vcl_name,
		     subject, set->vcl_name);
		return 0;
	}

	priv = VRT_priv_task(ctx, set);
	AN(priv);
	if (priv->priv == NULL) {
		if ((priv->priv = WS_Alloc(ctx->ws, sizeof(*task))) == NULL) {
			VERRNOMEM(ctx, ERR_PREFIX "allocating match data",
				  set->vcl_name, subject);
			return 0;
		}
		priv->len = sizeof(*task);
		priv->free = NULL;
		task = priv->priv;
		task->magic = TASK_SET_MATCH_MAGIC;
	}
	else {
		WS_Assert_Allocated(ctx->ws, priv->priv, sizeof(*task));
		CAST_OBJ(task, priv->priv, TASK_SET_MATCH_MAGIC);
	}

	buf = WS_Front(ctx->ws);
	buflen = WS_Reserve(ctx->ws, 0);
	if ((err = vre2set_match(set->set, subject, &match, buf, buflen,
				 &task->nmatches)) != NULL) {
		VERR(ctx, ERR_PREFIX "%s", set->vcl_name, subject, err);
		WS_Release(ctx->ws, 0);
		return 0;
	}
	if (match) {
		task->matches = (int *)buf;
		WS_Release(ctx->ws, task->nmatches * sizeof(int));
	}
	else
		WS_Release(ctx->ws, 0);
	return match;
}

#undef ERR_PREFIX

static struct task_set_match *
get_task_data(VRT_CTX, struct vmod_re2_set *set)
{
	struct vmod_priv *priv;
	struct task_set_match *task;

	priv = VRT_priv_task(ctx, set);
	AN(priv);
	if (priv->priv == NULL)
		return NULL;
	WS_Assert_Allocated(ctx->ws, priv->priv, sizeof(*task));
	CAST_OBJ(task, priv->priv, TASK_SET_MATCH_MAGIC);
	return task;
}

VCL_BOOL
vmod_set_matched(VRT_CTX, struct vmod_re2_set *set, VCL_INT n)
{
	struct task_set_match *task;
	int hi, lo = 0;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (n < 1 || n > set->npatterns) {
		VERR(ctx, "n=%d out of range in %s.matched() (%d patterns)", n,
		     set->vcl_name, set->npatterns);
		return 0;
	}

	if ((task = get_task_data(ctx, set)) == NULL) {
		VERR(ctx, "%s.matched(%d) called without prior match",
		     set->vcl_name, n);
		return 0;
	}

	if (task->nmatches == 0)
		return 0;
	WS_Assert_Allocated(ctx->ws, task->matches,
			    task->nmatches * sizeof(int));
	n--;
	hi = task->nmatches;
	do {
		int m = lo + (hi - lo) / 2;
		if (task->matches[m] == n)
			return 1;
		if (task->matches[m] < n)
			lo = m + 1;
		else
			hi = m - 1;
	} while (lo <= hi);

	return 0;
}

VCL_INT
vmod_set_nmatches(VRT_CTX, struct vmod_re2_set *set)
{
	struct task_set_match *task;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if ((task = get_task_data(ctx, set)) == NULL) {
		VERR(ctx, "%s.nmatches() called without prior match",
		     set->vcl_name);
		return 0;
	}
	return task->nmatches;
}

static int
get_match_idx(VRT_CTX, struct vmod_re2_set *set, VCL_INT n, VCL_ENUM selects,
	      const char *method)
{
	struct task_set_match *task;
	int idx = 0;

	if (n > set->npatterns) {
		VERR(ctx, "%s.%s(%lld): set has %d patterns", set->vcl_name,
		     method, n, set->npatterns);
		return -1;
	}
	if (n > 0)
		return n - 1;

	if ((task = get_task_data(ctx, set)) == NULL) {
		VERR(ctx, "%s.%s() called without prior match", set->vcl_name,
		     method);
		return -1;
	}

	if (task->nmatches == 0) {
		VERR(ctx, "%s.%s(%lld): previous match was unsuccessful",
		     set->vcl_name, method, n);
		return -1;
	}
	if (task->nmatches > 1) {
		if (selects == NULL) {
			VERR(ctx, "%s.%s(%lld): %d successful matches",
			     set->vcl_name, method, n, task->nmatches);
			return -1;
		}
		if (strcmp(selects, "LAST") == 0)
			idx = task->nmatches - 1;
		else
			AZ(strcmp(selects, "FIRST"));
	}
	WS_Assert_Allocated(ctx->ws, task->matches,
			    task->nmatches * sizeof(int));
	return task->matches[idx];
}

VCL_STRING
vmod_set_string(VRT_CTX, struct vmod_re2_set *set, VCL_INT n, VCL_ENUM selects)
{
	int idx;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (set->string == NULL) {
		VERR(ctx, "%s.string(%lld): No strings were set for %s",
		     set->vcl_name, n, set->vcl_name);
		return NULL;
	}

	idx = get_match_idx(ctx, set, n, selects, "string");
	if (idx < 0)
		return NULL;
	return set->string[idx];
}

VCL_BACKEND
vmod_set_backend(VRT_CTX, struct vmod_re2_set *set, VCL_INT n, VCL_ENUM selects)
{
	int idx;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(set, VMOD_RE2_SET_MAGIC);
	if (set->backend == NULL) {
		VERR(ctx, "%s.backend(%lld): No backends were set for %s",
		     set->vcl_name, n, set->vcl_name);
		return NULL;
	}

	idx = get_match_idx(ctx, set, n, selects, "backend");
	if (idx < 0)
		return NULL;
	return set->backend[idx];
}

/* Regex function interface */

#define ERR_PREFIX "re2.match(pattern=\"%.40s\", text=\"%.40s\"): "

VCL_BOOL
vmod_match(VRT_CTX, struct vmod_priv *priv, VCL_STRING pattern,
	   VCL_STRING subject, VCL_BOOL utf8, VCL_BOOL posix_syntax,
	   VCL_BOOL longest_match, VCL_INT max_mem, VCL_BOOL literal,
	   VCL_BOOL never_nl, VCL_BOOL dot_nl, VCL_BOOL never_capture,
	   VCL_BOOL case_sensitive, VCL_BOOL perl_classes,
	   VCL_BOOL word_boundary, VCL_BOOL one_line)
{
	vre2 *vre2 = NULL;
	task_match_t *task_match;
	const char *err;
	void *groups = NULL, **groupsp = &groups;
	int ngroups = 0;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (subject == NULL)
		subject = "";
	if (pattern == NULL) {
		VERR(ctx, ERR_PREFIX "pattern is undefined", "<undefined>",
		     subject);
		return 0;
	}

	if ((err = vre2_init(&vre2, pattern, utf8, posix_syntax, longest_match,
			     max_mem, literal, never_nl, dot_nl, never_capture,
			     case_sensitive, perl_classes, word_boundary,
			     one_line))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "Cannot compile: %s", pattern, subject,
		     err);
		vre2_fini(&vre2);
		return 0;
	}
	if (! never_capture) {
		if ((err = vre2_ngroups(vre2, &ngroups)) != NULL) {
			VERR(ctx, "Cannot obtain number of capturing groups: "
			     "%s", pattern, subject, err);
			vre2_fini(&vre2);
			return 0;
		}
		assert(ngroups >= 0);
	}
	if (priv->priv == NULL) {
		if ((task_match = WS_Alloc(ctx->ws, sizeof(*task_match)))
		    == NULL) {
			VERRNOMEM(ctx, ERR_PREFIX "allocating match data",
				  pattern, subject);
			vre2_fini(&vre2);
			return 0;
		}
		priv->priv = task_match;
		priv->len = sizeof(*task_match);
		priv->free = free_task_match;
		task_match->magic = TASK_MATCH_MAGIC;
	}
	else {
		WS_Assert_Allocated(ctx->ws, priv->priv, sizeof(*task_match));
		CAST_OBJ(task_match, priv->priv, TASK_MATCH_MAGIC);
	}
	task_match->ngroups = ngroups;
	task_match->never_capture = never_capture;
	task_match->vre2 = vre2;
	groupsp = &task_match->groups;

	return match(ctx, vre2, subject, groupsp, never_capture, ngroups);
}

#undef ERR_PREFIX

#define ERR_PREFIX "re2.backref(ref=%ld, fallback=\"%.40s\"): "

VCL_STRING
vmod_backref(VRT_CTX, struct vmod_priv *priv, VCL_INT refnum,
	     VCL_STRING fallback)
{
	task_match_t *task_match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);
	assert(refnum >= 0);
	if (fallback == NULL) {
		VERR(ctx, ERR_PREFIX "fallback is undefined",  refnum,
		     "<undefined>");
		return "**BACKREF FUNCTION FAILED**";
	}
	if (priv->priv == NULL) {
		VERR(ctx, ERR_PREFIX "called without previous match", refnum,
		     fallback);
		return fallback;
	}

	WS_Assert_Allocated(ctx->ws, priv->priv, sizeof(*task_match));
	CAST_OBJ(task_match, priv->priv, TASK_MATCH_MAGIC);

	if (task_match->never_capture) {
		VERR(ctx, ERR_PREFIX "never_capture was true in previous match",
		     refnum, fallback);
		return fallback;
	}
	if (refnum > task_match->ngroups) {
		VERR(ctx, ERR_PREFIX "backref out of range (max %d)", refnum,
		     fallback, task_match->ngroups);
		return fallback;
	}
	return backref(ctx, refnum, fallback, task_match->groups,
		       task_match->ngroups);
}

#undef ERR_PREFIX

#define ERR_PREFIX "re2.namedref(name=\"%.40s\", fallback=\"%.40s\"): "

VCL_STRING
vmod_namedref(VRT_CTX, struct vmod_priv *priv, VCL_STRING name,
	      VCL_STRING fallback)
{
	task_match_t *task_match;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);
	if (fallback == NULL) {
		ERR(ctx, "re2.namedref(): fallback is undefined");
		return "**NAMEDREF FUNCTION FAILED**";
	}
	if (name == NULL || name[0] == '\0') {
		VERR(ctx, ERR_PREFIX "name is empty", "", fallback);
		return fallback;
	}
	if (priv->priv == NULL) {
		VERR(ctx, ERR_PREFIX "called without previous match", name,
		     fallback);
		return fallback;
	}

	WS_Assert_Allocated(ctx->ws, priv->priv, sizeof(*task_match));
	CAST_OBJ(task_match, priv->priv, TASK_MATCH_MAGIC);

	if (task_match->never_capture) {
		VERR(ctx, ERR_PREFIX "never_capture was true in previous match",
		     name, fallback);
		return fallback;
	}
	return namedref(ctx, task_match->vre2, name, fallback,
			task_match->groups, task_match->ngroups);
}

#undef ERR_PREFIX

#define ERR_PREFIX "re2.%s(pattern=\"%.40s\", text=\"%.40s\", rewrite=\"%.40s\", fallback=\"%.40s\"): "

static VCL_STRING
rewrite_function(VRT_CTX, const rewrite_e mode, VCL_STRING restrict pattern,
		 VCL_STRING restrict text, VCL_STRING restrict rewrite,
		 VCL_STRING restrict fallback, VCL_BOOL utf8,
		 VCL_BOOL posix_syntax, VCL_BOOL longest_match, VCL_INT max_mem,
		 VCL_BOOL literal, VCL_BOOL never_nl, VCL_BOOL dot_nl,
		 VCL_BOOL never_capture, VCL_BOOL case_sensitive,
		 VCL_BOOL perl_classes, VCL_BOOL word_boundary,
		 VCL_BOOL one_line)
{
	vre2 *vre2 = NULL;
	const char *ret, *err;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (fallback == NULL) {
		VERR(ctx, "re2.%s(): fallback is undefined",
		     rewrite_name[mode]);
		switch(mode) {
		case SUB:
			return "**SUB FUNCTION FAILED**";
		case SUBALL:
			return "**SUBALL FUNCTION FAILED**";
		case EXTRACT:
			return "**EXTRACT FUNCTION FAILED**";
		default:
			WRONG("illegal mode");
		}
	}
	if (pattern == NULL) {
		VERR(ctx, "re2.%s(pattern=<undefined>, fallback=\"%.40s\"): "
		     "pattern is undefined", rewrite_name[mode], fallback);
		return fallback;
	}
	if (text == NULL) {
		VERR(ctx, "re2.%s(pattern=\"%s\", text=<undefined>, "
		     "fallback=\"%.40s\"): text is undefined",
		     rewrite_name[mode], pattern, fallback);
		return fallback;
	}
	if (rewrite == NULL) {
		VERR(ctx, "re2.%s(pattern=\"%.40s\", text=\"%.40s\", "
		     "rewrite=<undefined>, fallback=\"%.40s\"): "
		     "rewrite is undefined", rewrite_name[mode], pattern, text,
		     fallback);
		return fallback;
	}

	if ((err = vre2_init(&vre2, pattern, utf8, posix_syntax, longest_match,
			     max_mem, literal, never_nl, dot_nl, never_capture,
			     case_sensitive, perl_classes, word_boundary,
			     one_line))
	    != NULL) {
		VERR(ctx, ERR_PREFIX "Cannot compile '%s': %s",
		     rewrite_name[mode], pattern, text, rewrite, fallback,
		     pattern, err);
		vre2_fini(&vre2);
		return fallback;
	}
	ret = rewritef(ctx, vre2, mode, text, rewrite, fallback);
	vre2_fini(&vre2);
	return ret;
}

#undef ERR_PREFIX

VCL_STRING
vmod_sub(VRT_CTX, VCL_STRING pattern, VCL_STRING text, VCL_STRING rewrite,
	 VCL_STRING fallback, VCL_BOOL utf8, VCL_BOOL posix_syntax,
	 VCL_BOOL longest_match, VCL_INT max_mem, VCL_BOOL literal,
	 VCL_BOOL never_nl, VCL_BOOL dot_nl, VCL_BOOL never_capture,
	 VCL_BOOL case_sensitive, VCL_BOOL perl_classes, VCL_BOOL word_boundary,
	 VCL_BOOL one_line)
{
	return rewrite_function(ctx, SUB, pattern, text, rewrite, fallback,
				utf8, posix_syntax, longest_match, max_mem,
				literal, never_nl, dot_nl, never_capture,
				case_sensitive, perl_classes, word_boundary,
				one_line);
}

VCL_STRING
vmod_suball(VRT_CTX, VCL_STRING pattern, VCL_STRING text, VCL_STRING rewrite,
	    VCL_STRING fallback, VCL_BOOL utf8, VCL_BOOL posix_syntax,
	    VCL_BOOL longest_match, VCL_INT max_mem, VCL_BOOL literal,
	    VCL_BOOL never_nl, VCL_BOOL dot_nl, VCL_BOOL never_capture,
	    VCL_BOOL case_sensitive, VCL_BOOL perl_classes,
	    VCL_BOOL word_boundary, VCL_BOOL one_line)
{
	return rewrite_function(ctx, SUBALL, pattern, text, rewrite, fallback,
				utf8, posix_syntax, longest_match, max_mem,
				literal, never_nl, dot_nl, never_capture,
				case_sensitive, perl_classes, word_boundary,
				one_line);
}

VCL_STRING
vmod_extract(VRT_CTX, VCL_STRING pattern, VCL_STRING text, VCL_STRING rewrite,
	     VCL_STRING fallback, VCL_BOOL utf8, VCL_BOOL posix_syntax,
	     VCL_BOOL longest_match, VCL_INT max_mem, VCL_BOOL literal,
	     VCL_BOOL never_nl, VCL_BOOL dot_nl, VCL_BOOL never_capture,
	     VCL_BOOL case_sensitive, VCL_BOOL perl_classes,
	     VCL_BOOL word_boundary, VCL_BOOL one_line)
{
	return rewrite_function(ctx, EXTRACT, pattern, text, rewrite, fallback,
				utf8, posix_syntax, longest_match, max_mem,
				literal, never_nl, dot_nl, never_capture,
				case_sensitive, perl_classes, word_boundary,
				one_line);
}

VCL_STRING
vmod_version(const struct vrt_ctx *ctx __attribute__((unused)))
{
	return VERSION;
}
